@@ -124,6 +124,7 @@ struct _MenuCache
     GSList* notifiers;
     GThread* thr;
     GCancellable* cancellable;
+    guint version;
     gboolean ready : 1; /* used for sync access */
 };
 
@@ -631,6 +632,7 @@ gboolean menu_cache_reload( MenuCache* cache )
     GDataInputStream* f;
     MenuCacheFileDir** all_used_files;
     int i, n;
+    int ver_maj, ver_min;
 
     file = g_file_new_for_path(cache->cache_file);
     if(!file)
@@ -648,12 +650,12 @@ gboolean menu_cache_reload( MenuCache* cache )
     line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
     if(G_LIKELY(line))
     {
-        int ver_maj, ver_min;
         len = sscanf(line, "%d.%d", &ver_maj, &ver_min);
         g_free(line);
         if(len < 2)
             goto _fail;
-        if( ver_maj != VER_MAJOR || ver_min != VER_MINOR )
+        if( ver_maj != VER_MAJOR ||
+            ver_min > VER_MINOR || ver_min < VER_MINOR_SUPPORTED )
             goto _fail;
     }
     else
@@ -695,6 +697,7 @@ _fail:
         g_object_unref(f);
         return FALSE;
     }
+    cache->version = ver_min;
 
     if(cache->root_dir)
         menu_cache_item_unref( MENU_CACHE_ITEM(cache->root_dir) );
@@ -1349,7 +1352,7 @@ static void get_socket_name( char* buf, int len )
 
 #define MAX_RETRIES 25
 
-static gboolean fork_server()
+static gboolean fork_server(const char *path)
 {
     int ret, pid, status;
 
@@ -1362,8 +1365,9 @@ static gboolean fork_server()
     pid = fork();
     if (pid == 0)
     {
-        execl( MENUCACHE_LIBEXECDIR "/menu-cached", MENUCACHE_LIBEXECDIR "/menu-cached", NULL);
-        g_print("failed to exec %s\n", MENUCACHE_LIBEXECDIR "/menu-cached");
+        execl(MENUCACHE_LIBEXECDIR "/menu-cached", MENUCACHE_LIBEXECDIR "/menu-cached",
+              path, NULL);
+        g_print("failed to exec %s %s\n", MENUCACHE_LIBEXECDIR "/menu-cached", path);
     }
 
     /*
@@ -1378,24 +1382,27 @@ retry_wait:
     return TRUE;
 }
 
-static gpointer server_io_thread(gpointer _unused)
+/* this thread is started by connect_server() */
+static gpointer server_io_thread(gpointer data)
 {
     char buf[1024]; /* protocol has a lot shorter strings */
     ssize_t sz;
     size_t ptr = 0;
-    int fd;
+    int fd = GPOINTER_TO_INT(data);
     GHashTableIter it;
     char* menu_name;
     MenuCache* cache;
 
-    G_LOCK(connect);
-    fd = server_fd;
-    G_UNLOCK(connect);
     while(fd >= 0)
     {
         sz = read(fd, &buf[ptr], sizeof(buf) - ptr);
         if(sz <= 0) /* socket error or EOF */
         {
+            MENU_CACHE_LOCK;
+            ptr = hash ? g_hash_table_size(hash) : 0;
+            MENU_CACHE_UNLOCK;
+            if (ptr == 0) /* don't need it anymore */
+                break;
             G_LOCK(connect);
             if(fd != server_fd) /* someone replaced us?! go out immediately! */
             {
@@ -1405,9 +1412,10 @@ static gpointer server_io_thread(gpointer _unused)
             server_fd = -1;
             G_UNLOCK(connect);
             DEBUG("connect failed, trying reconnect");
+            sleep(1);
             if( ! connect_server(NULL) )
             {
-                g_print("fail to re-connect to the server.\n");
+                g_critical("fail to re-connect to the server.");
                 MENU_CACHE_LOCK;
                 if(hash)
                 {
@@ -1418,7 +1426,7 @@ static gpointer server_io_thread(gpointer _unused)
                 MENU_CACHE_UNLOCK;
                 break;
             }
-            DEBUG("successfully restart server.\nre-register menus.");
+            DEBUG("successfully reconnected server, re-register menus.");
             /* re-register all menu caches */
             MENU_CACHE_LOCK;
             if(hash)
@@ -1429,7 +1437,7 @@ static gpointer server_io_thread(gpointer _unused)
                     /* FIXME: need we remove it from hash if failed? */
             }
             MENU_CACHE_UNLOCK;
-            continue;
+            break; /* next thread will do it */
         }
         while(sz > 0)
         {
@@ -1480,6 +1488,11 @@ static gpointer server_io_thread(gpointer _unused)
             ptr = 0;
         }
     }
+    G_LOCK(connect);
+    if (fd == server_fd)
+        server_fd = -1;
+    G_UNLOCK(connect);
+    close(fd);
     /* DEBUG("server io thread terminated"); */
 #if GLIB_CHECK_VERSION(2, 32, 0)
     g_thread_unref(g_thread_self());
@@ -1525,7 +1538,7 @@ retry:
         if((rc == ECONNREFUSED || rc == ENOENT) && retries == 0)
         {
             DEBUG("no running server found, starting it");
-            fork_server();
+            fork_server(addr.sun_path);
             ++retries;
             goto retry;
         }
@@ -1542,13 +1555,17 @@ retry:
     server_fd = fd;
     G_UNLOCK(connect);
 #if GLIB_CHECK_VERSION(2, 32, 0)
-    g_thread_new("menu-cache-io", server_io_thread, NULL);
+    g_thread_new("menu-cache-io", server_io_thread, GINT_TO_POINTER(fd));
 #else
-    g_thread_create(server_io_thread, NULL, FALSE, NULL);
+    g_thread_create(server_io_thread, GINT_TO_POINTER(fd), FALSE, NULL);
 #endif
     return TRUE;
 }
 
+#define CACHE_VERSION __num2str(VER_MAJOR) "." __num2str(VER_MINOR)
+#define __num2str(s) __def2str(s)
+#define __def2str(s) #s
+
 static MenuCache* menu_cache_create(const char* menu_name)
 {
     MenuCache* cache;
@@ -1564,6 +1581,7 @@ static MenuCache* menu_cache_create(const char* menu_name)
     char* file_name;
     int len = 0;
     GChecksum *sum;
+    char *langs_list;
 
     if( !xdg_cfg )
         xdg_cfg = "";
@@ -1578,13 +1596,13 @@ static MenuCache* menu_cache_create(const char* menu_name)
     if( ! xdg_cache_home )
         xdg_cache_home = "";
 
-    /* get rid of the encoding part of locale name. */
-    while( strchr(langs[0], '.') )
-        ++langs;
+    /* reconstruct languages list in form as it should be in $LANGUAGES */
+    langs_list = g_strjoinv(":", (char **)langs);
 
-    buf = g_strdup_printf( "REG:%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t00000000000000000000000000000000\n",
+    buf = g_strdup_printf( "REG:%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t" CACHE_VERSION
+                           "\t00000000000000000000000000000000\n",
                             menu_name,
-                            *langs,
+                            langs_list,
                             xdg_cache_home,
                             xdg_cfg,
                             xdg_prefix,
@@ -1605,6 +1623,7 @@ static MenuCache* menu_cache_create(const char* menu_name)
     memcpy( cache->md5, md5, 32 );
     cache->menu_name = g_strdup(menu_name);
     g_free( file_name );
+    g_free(langs_list);
 
     g_checksum_free(sum); /* md5 is also freed here */
 
@@ -1650,9 +1669,8 @@ static gpointer menu_cache_loader_thread(gpointer data)
         return NULL;
     }
     /* and request update from server */
-    if(!cache->cancellable || !g_cancellable_is_cancelled(cache->cancellable))
-        register_menu_to_server(cache);
-    else
+    if ((cache->cancellable && g_cancellable_is_cancelled(cache->cancellable)) ||
+        !register_menu_to_server(cache))
         SET_CACHE_READY(cache);
     return NULL;
 }
@@ -1899,3 +1917,6 @@ MenuCacheItem *menu_cache_find_item_by_id(MenuCache *cache, const char *id)
     MENU_CACHE_UNLOCK;
     return item;
 }
+
+/* TODO:
+const char **menu_cache_app_get_categories(MenuCacheApp *app) */
\ No newline at end of file
