@@ -73,10 +73,12 @@ typedef struct _Cache
     char *cache_file;
 }Cache;
 
-GMainLoop* main_loop = NULL;
+static GMainLoop* main_loop = NULL;
 
 static GHashTable* hash = NULL;
 
+static char *socket_file = NULL;
+
 static void on_file_changed( GFileMonitor* mon, GFile* gf, GFile* other,
                              GFileMonitorEvent evt, Cache* cache );
 
@@ -181,7 +183,7 @@ static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
 
 static void set_env( char* penv, const char* name )
 {
-    if( *penv )
+    if( penv && *penv )
         g_setenv( name, penv, TRUE);
     else
         g_unsetenv( name );
@@ -196,6 +198,7 @@ static void pre_exec( gpointer user_data )
     set_env(*++env, "XDG_DATA_DIRS");
     set_env(*++env, "XDG_CONFIG_HOME");
     set_env(*++env, "XDG_DATA_HOME");
+    set_env(*++env, "CACHE_GEN_VERSION");
 }
 
 static gboolean regenerate_cache( const char* menu_name,
@@ -440,10 +443,9 @@ static void get_socket_name( char* buf, int len )
     g_free(dpy);
 }
 
-static int create_socket()
+static int create_socket(struct sockaddr_un *addr)
 {
     int fd = -1;
-    struct sockaddr_un addr;
 
     fd = socket(PF_UNIX, SOCK_STREAM, 0);
     if (fd < 0)
@@ -451,20 +453,17 @@ static int create_socket()
         DEBUG("Failed to create socket");
         return -1;
     }
-    memset(&addr, 0, sizeof(addr));
-    addr.sun_family = AF_UNIX;
-    get_socket_name( addr.sun_path, sizeof( addr.sun_path ) );
 
     /* remove previous socket file */
-    if (unlink(addr.sun_path) < 0) {
+    if (unlink(addr->sun_path) < 0) {
         if (errno != ENOENT)
-            g_error("Couldn't remove previous socket file %s", addr.sun_path);
+            g_error("Couldn't remove previous socket file %s", addr->sun_path);
     }
     /* remove of previous socket file successful */
     else
-        g_warning("removed previous socket file %s", addr.sun_path);
+        g_warning("removed previous socket file %s", addr->sun_path);
 
-    if(bind(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0)
+    if(bind(fd, (struct sockaddr *)addr, sizeof(*addr)) < 0)
     {
         DEBUG("Failed to bind to socket");
         close(fd);
@@ -539,10 +538,10 @@ retry:
 
     if( memcmp(line, "REG:", 4) == 0 )
     {
-        int n_files, i;
+        int n_files = 0, i;
         char *pline = line + 4;
         char *sep, *menu_name, *lang_name, *cache_dir;
-        char **files;
+        char **files = NULL;
         char **env;
         char reload_cmd[38] = "REL:";
 
@@ -555,7 +554,9 @@ retry:
          * XDG_MENU_PREFIX
          * XDG_DATA_DIRS
          * XDG_CONFIG_HOME
-         * XDG_DATA_HOME */
+         * XDG_DATA_HOME
+         * (optional) CACHE_GEN_VERSION
+         * md5 hash */
 
         md5 = pline + len - 32; /* the md5 hash of this menu */
 
@@ -572,6 +573,7 @@ retry:
             lang_name = pline;
             pline = sep + 1;
 
+            ((char *)md5)[-1] = '\0';
             env = g_strsplit(pline, "\t", 0);
 
             cache_dir = env[0]; /* XDG_CACHE_HOME */
@@ -695,9 +697,7 @@ static gboolean on_new_conn_incoming(GIOChannel* ch, GIOCondition cond, gpointer
 static void terminate(int sig)
 {
 /* #ifndef HAVE_ABSTRACT_SOCKETS */
-    char path[1024];
-    get_socket_name(path, sizeof(path));
-    unlink(path);
+    unlink(socket_file);
     exit(0);
 /* #endif */
 }
@@ -710,11 +710,11 @@ static gboolean on_server_conn_close(GIOChannel* ch, GIOCondition cond, gpointer
     return TRUE;
 }
 
-
 int main(int argc, char** argv)
 {
     GIOChannel* ch;
     int server_fd;
+    struct sockaddr_un addr;
 #ifndef DISABLE_DAEMONIZE
     int fd, pid;
 
@@ -722,7 +722,21 @@ int main(int argc, char** argv)
     long i;
 #endif
 
-    server_fd = create_socket();
+    memset(&addr, 0, sizeof(addr));
+    addr.sun_family = AF_UNIX;
+    if (argc < 2)
+        /* old way, not recommended! */
+        get_socket_name( addr.sun_path, sizeof( addr.sun_path ) );
+    else if (strlen(argv[1]) >= sizeof(addr.sun_path))
+        /* ouch, it's too big! */
+        return 1;
+    else
+        /* this is a good way! */
+        strncpy(addr.sun_path, argv[1], sizeof(addr.sun_path) - 1);
+
+    socket_file = addr.sun_path;
+
+    server_fd = create_socket(&addr);
 
     if( server_fd < 0 )
         return 1;
@@ -794,11 +808,7 @@ int main(int argc, char** argv)
     g_main_loop_run( main_loop );
     g_main_loop_unref( main_loop );
 
-    {
-        char path[256];
-        get_socket_name(path, 256 );
-        unlink(path);
-    }
+    unlink(addr.sun_path);
 
     g_hash_table_destroy(hash);
     return 0;
